import os
import random

def get_in_drive(path: str) -> bytes:
	with open(path, 'rb') as _buffer:
		return _buffer.read()

def merge_permissions(d1: dict[str, str], d2: dict[str, str]) -> dict[str, str]:
	new_dict: dict[str, str] = d1.copy()

	for key, val in d2.items():
		new_dict[key] = bool(val or new_dict.get(key, False))

	return new_dict

def shorten(text: str, length: int, dots: bool = True) -> str:
	if dots:
		return text[:length - 3] + "..."
	else:
		return text[:length]


# Sécurité

def adjust_path(path: str) -> str:
	path = os.path.normpath(path)
	path = path.replace('\\', '/')

	newpath = []
	for component in path.split('/'):
		if component not in ('..',):
			newpath.append(component)

	return '/'.join(newpath)


# Banque

def gen_digicode(length: int = 6) -> str:
	return hex(random.randint(0, 16 ** length - 1))[2:].upper().zfill(length)

# Impôts

TAXATIONS = {
	'taxe_ega': "105",
	'taxe_dc': "106"
}

BONUSES = {
	'aide_ega': "105",
	'benevolat': "105",
	'campagnes': "6"
}

def calculate_amount(amount: int = 0, specs: dict = {},history: dict = {}) -> dict:
	sheet = {
		"aide_ega": 0,
		"benevolat": 0,
		"campagnes": 0,
		"taxe_ega": 0,
		"taxe_dc": 0
	}

	# Taxe égalitaire
	if amount < 2000:
		percentage = 0
	elif 2000 <= amount < 5000:
		percentage = .01
	elif 5000 <= amount < 20000:
		percentage = .025
	elif 20000 <= amount < 50000:
		percentage = .045
	else:
		percentage = .05
		e = amount

		while e >= 150000 and percentage < .2:
			percentage += .01
			e -= 100000

	sheet['taxe_ega'] += int(round(amount * percentage))

	# Taxe double compte
	if specs.get('DC'):
		if amount < 2000:
			percentage = .01
		elif 2000 <= amount < 5000:
			percentage = .025
		elif 5000 <= amount < 20000:
			percentage = .045
		elif 20000 <= amount < 50000:
			percentage = .05
		else:
			percentage = .1
			e = amount

			while e >= 150000 and percentage < .3:
				percentage += .05
				e -= 100000

		sheet['taxe_dc'] += int(round(amount * percentage))


	"""
	AIDES & BONUS
	"""

	# Aide égalitaire
	if amount < 2000:
		sheet['aide_ega'] -= 2000 - amount

	# Bonus de dons aux assos et de participation aux campagnes
	if amount != 0:
		dons = history.get('dons', 0)

		if dons / amount < .05:
			sheet['benevolat'] -= .1 * dons
		elif .05 <= dons / amount < .1:
			sheet['benevolat'] -= .25 * dons
		elif .1 <= dons / amount < .25:
			sheet['benevolat'] -= .4 * dons
		elif .25 <= dons / amount < .4:
			sheet['benevolat'] -= .6 * dons
		elif .4 <= dons / amount < .6:
			sheet['benevolat'] -= .75 * dons
		elif .6 <= dons / amount < 1:
			sheet['benevolat'] -= dons

	campagnes = history.get('campagnes', 0)

	if campagnes >= 0:
		sheet['campagnes'] -= campagnes

	return sheet


"""
nsarchive - API-wrapper pour récupérer des données liées à Nation.

Version: 4.4.2
License: GPL-3.0
Auteur : happex <110610727+okayhappex@users.noreply.github.com>

Dependencies:
- Python ^3.10
- pillow ^10.4

Le fichier README.md fournit des détails supplémentaires pour l'utilisation.
"""

# Import des types 
from .models.base import NSID
from .models.entities import *
from .models.economy import *

from .models.state import *
from .models.justice import *

from .models.scale import *

# Import des interfaces
from .models.base import Interface
from .interfaces._entities import EntityInterface
from .interfaces._economy import EconomyInterface
from .interfaces._state import StateInterface
from .interfaces._justice import JusticeInterface


import time

from .base import NSID

from .. import database as db

class Report:
    def __init__(self, id: NSID):
        self._path: str = ''

        self.id: NSID = id
        self.author: NSID = NSID('0')
        self.target: NSID = NSID('0')
        self.date: int = round(time.time())
        self.status: int = 0 # 0: En attente, 1: Accepté, 2: Rejeté
        self.reason: str = None # Raison proposée par le bot
        self.details:str = None # Description des faits

    def _load(self, _data: dict, path: str) -> None:
        self._path = path

        self.id = NSID(_data['id'])
        self.author = NSID(_data['author'])
        self.target = NSID(_data['target'])
        self.date = _data['date']
        self.status = _data['status']
        self.reason = _data.get('reason', None)
        self.details = _data.get('details', None)

    def _to_dict(self) -> dict:
        return {
            'id': self.id,
            'target': self.target,
            'author': self.author,
            'date': self.date,
            'status': self.status,
            'reason': self.reason,
            'details': self.details
        }

    def save(self):
        db.put_item(self._path, 'reports', self._to_dict())

    def update(self, status: str | int):
        __statuses = [
            'pending',
            'accepted',
            'rejected'
        ]

        if status not in __statuses:
            if isinstance(status, int) and 0 <= status <= 2:
                status = __statuses[status]
            else:
                raise ValueError(f"Invalid status: {status}. Must be one of {__statuses} or an integer between 0 and 2.")

        self.status = __statuses.index(status)
        self.save()

class Sanction:
    def __init__(self, id: NSID):
        self._path: str = ''

        self.id: NSID = id
        self.target: NSID = NSID('0')
        self.type: str = None
        self.date: int = round(time.time())
        self.duration: int = 0
        self.title: str = None
        self.lawsuit: Lawsuit = None

    def _load(self, _data: dict, path: str,) -> None:
        self._path = path

        self.id = NSID(_data['id'])
        self.target = NSID(_data['target'])
        self.type = _data['type']
        self.date = _data['date']
        self.duration = _data['duration']
        self.title = _data['title']

        lawsuit = db.get_item(path, 'lawsuits', _data['lawsuit'])

        if lawsuit:
            self.lawsuit = Lawsuit(NSID(lawsuit))
            self.lawsuit._load(lawsuit, path)

    def _to_dict(self) -> dict:
        return {
            'id': self.id,
            'target': self.target,
            'type': self.type,
            'date': self.date,
            'duration': self.duration,
            'title': self.title,
            'lawsuit': self.lawsuit if self.lawsuit else None
        }

    def save(self):
        db.put_item(self._path, 'sanctions', self._to_dict())

class Lawsuit:
    def __init__(self, id: NSID):
        self._path: str = ''

        self.id: NSID = id
        self.target: NSID = NSID('0')
        self.judge: NSID = NSID('0')
        self.title: str = None
        self.date: int = round(time.time())
        self.report: Report = None
        self.is_private: bool = False
        self.is_open: bool = False

    def _load(self, _data: dict, path: str) -> None:
        self._path = path

        self.id = NSID(_data['id'])
        self.target = NSID(_data['target'])
        self.judge = NSID(_data['judge'])
        self.title = _data.get('title')
        self.date = _data.get('date', round(time.time()))

        report = db.get_item(path, 'reports', _data['report'])

        if report: 
            self.report = Report(NSID(report))
            self.report._load(report, path)

        self.is_private = bool(_data.get('private', 0))
        self.is_open = _data.get('status', 0) == 0

    def _to_dict(self) -> dict:
        return {
            'id': self.id,
            'target': self.target,
            'judge': self.judge,
            'title': self.title,
            'date': self.date,
            'report': self.report.id if self.report else None,
            'is_private': self.is_private,
            'is_open': self.is_open
        }

    def save(self):
        db.put_item(self._path, 'reports', self._to_dict())


import os
import time
import typing

from ..models.base import NSID, Interface
from ..models.entities import _load_position, Entity, User, Organization, Position, PositionPermissions, Certification

from .. import database as db


class EntityInterface(Interface):
    """
    Interface qui vous permettra d'interagir avec les profils des membres ainsi que les différents métiers et secteurs d'activité.

    ## Informations disponibles
    - Profil des membres et des entreprises: `.User | .Organization`
    - Appartenance et permissions d'un membre dans un groupe: `.GroupMember.MemberPermissions`
    - Position légale et permissions d'une entité: `.Position.Permissions`
    """

    def __init__(self, path: str) -> None:
        super().__init__(os.path.join(path, 'entities'))

    """
    ---- ENTITÉS ----
    """

    def get_entity(self, id: NSID, _class: typing.Type) -> User | Organization | None:
        """
        Fonction permettant de récupérer le profil public d'une entité.\n

        ## Paramètres
        - id: `NSID`\n
            ID héxadécimal de l'entité à récupérer
        - _class: `Type`\n
            Classe du modèle à prendre (`.User` ou `.Organization`)

        ## Renvoie
        - `.User` dans le cas où l'entité choisie est un membre
        - `.Organization` dans le cas où c'est un groupe
        - `None` dans le cas où c'est indéterminé ou l'entité n'existe pas
        """

        id = NSID(id)

        if _class == User:
            data = db.get_item(self.path, 'individuals', id)
        elif _class == Organization:
            data = db.get_item(self.path, 'organizations', id)
        elif _class is None:
            data = db.get_item(self.path, 'individuals', id)
            _class = User

            if not data:
                data = db.get_item(self.path, 'organizations', id)
                _class = Organization

            if not data:
                _class = None
        else:
            raise ValueError(f"Invalid class type: {_class.__name__}")

        if not data:
            return


        # TRAITEMENT

        if _class == User:
            entity = User(id)
        elif _class == Organization:
            entity = Organization(id)
        else:
            return


        entity._load(data, self.path)

        return entity

    def create_entity(self, id: NSID, name: str, _class: typing.Type, position: str = None):
        """
        Fonction permettant de créer ou modifier une entité.

        ## Paramètres
        - id (`NSID`): Identifiant NSID
        - name (`str`): Nom d'usage
        - _class (`.User` ou `.Organization`): Type de l'entité
        - position (`str`, optionnel): ID de la position civile
        """

        id = NSID(id)

        if _class == Organization:
            data = {
                'id': id,
                'name': name,
                'position': position if position else 'group',
                'register_date': round(time.time()),
                'owner_id': NSID(0x100), # self.session.author, | TODO: Implémenter les Sessions
                'certifications': {}, # Implémenter les Certifications
                'members': {},
                'additional': {}
            }
        elif _class == User:
            data = {
                'id': id,
                'name': name,
                'position': position if position else 'member',
                'register_date': round(time.time()),
                'certifications': {},
                'xp': 0,
                'boosts': {},
                'additional': {}
            }
        else:
            raise ValueError(f"Class '{_class.__name__}' not recognized.")

        db.put_item(self.path, 'individuals' if _class == User else 'organizations', data, True)


        # TRAITEMENT

        entity = User(id) if _class == User else Organization(id)
        entity._load(data, self.path)

        return entity

    def delete_entity(self, id: NSID):
        """
        Fonction permettant de supprimer le profil d'une entité

        ## Paramètres
        id: `NSID`\n
            L'ID de l'entité à supprimer
        """

        try:
            db.delete_item(self.path, 'individuals', NSID(id))
        except KeyError:
            db.delete_item(self.path, 'organizations', NSID(id))

    def fetch_entities(self, _class: typing.Type, **query: typing.Any) -> list[User] | list[Organization]:
        """
        Récupère une liste d'entités en fonction d'une requête.

        ## Paramètres
        - _class (`.User` ou `.Organization`):\n
            Table dans laquelle chercher les utilisateurs
        - query: `**dict`\n
            La requête pour filtrer les entités.

        ## Renvoie
        - `list[.User | .Organization]`
        """

        if _class == User:
            table = 'individuals'
        elif _class == Organization:
            table = 'organizations'
        else:
            raise ValueError(f"Class '{_class.__name__}' is not recognized.")

        _res = db.fetch(f"{self.path}:{table}", **query)


        res = []

        for _entity in _res:
            if _entity is None: continue

            if _class == User:
                entity = User(_entity["id"])
            elif _class == Organization:
                entity = Organization(_entity["id"])
            else:
                entity = Entity(_entity["id"])

            entity._load(_entity, self.path)

            res.append(entity)

        return res



    def get_position(self, id: str) -> Position:
        """
        Récupère une position légale (métier, domaine professionnel).

        ## Paramètres
        id: `str`\n
            ID de la position (SENSIBLE À LA CASSE !)

        ## Renvoie
        - `.Position`
        """

        data = _load_position(id, self.path)

        if not data:
            return

        # TRAITEMENT

        position = Position(id)
        pos = _load_position(id, self.path)
        position._load(pos, self.path)

        return position

    def get_position_tree(self, id: str, tree: tuple = ()) -> tuple:
        position = self.get_position(id)

        if position.root:
            return self.get_position_tree(position.root, tree + position.id,)
        else:
            return tree + position.id,

    def permission_herit(self, id: str, permissions: PositionPermissions = PositionPermissions()) -> tuple:
        position = self.get_position(id)

        permissions.merge(position.permissions)

        if position.root:
            self.permission_herit(position.root, permissions)

    def create_position(
        self,
        id: str,
        title: str,
        permissions: PositionPermissions = PositionPermissions(),
        root: Position = None,
        level: int = None
    ) -> Position:
        """
        Crée une position légale

        ## Paramètres
        - id: `str`\n
            ID de la position
        - title: `str`\n
            Titre de la position
        - root: `str`\n
            Catégorie de la position (officier, citoyen)
        - level: `int`\n
            Grade de la position dans cette catégorie
        - permissions: `.PositionPermissions` (optionnel)\n
            Permissions accordées à la position
        """

        if root and not level:
            raise ValueError("If 'root' is specified, 'level' must also be specified.")

        data = {
            'id': id,
            'name': title,
            'role': None,
            'root': root.id if root else None,
            'level': level,
            'permissions': permissions.__dict__
        }

        db.put_item(self.path, 'positions', data)


        position = Position(id)
        pos = _load_position(id, self.path)
        position._load(pos, self.path)

        return position

    def delete_position(self, id: str):
        db.delete_item(self.path, 'positions', id)

    def fetch_positions(self, **query: typing.Any) -> list[Position]:
        """
        Récupère une liste de positions en fonction d'une requête.

        ## Paramètres
        query: `**dict`\n
            La requête pour filtrer les positions.

        ## Renvoie
        - `list[.Position]`
        """

        _res = db.fetch('positions', **query)
        res = []

        for _data in _res:
            pos = Position()
            pos._load(_data, self.path)

            self.permission_herit(pos.id, pos.permissions)

            res.append(pos)

        return res


    def get_certification(self, id: str) -> Certification:
        """
        Récupère une certification.

        ## Paramètres
        id: `str`\n
            ID de la certification (SENSIBLE À LA CASSE !)

        ## Renvoie
        - `.Certification`
        """

        data = db.get_item(self.path, 'certifications', id)

        if not data:
            return

        # TRAITEMENT

        certification = Certification(id)
        certification._load(data, self.path)

        return certification

    def register_certification(self, id: str, title: str, owner: Organization, parent: Certification = None, duration: int = 2419200) -> Certification:
        """
        Crée une certification

        ## Paramètres
        - id: `str`\n
            ID de la certification
        - title: `str`\n
            Titre de la certification
        - owner: `Organization`\n
            Organisation propriétaire de la certification
        - parent: `Certification`\n
            Certification mère (nécessaire à l'entité qui voudra délivrer celle-ci)
        - duration: `int`\n
            Durée de la certification en secondes (par défaut 28 jours)
        """

        data = {
            'id': id,
            'name': title,
            'duration': duration,
            'parent': parent.id if parent else None,
            'owner': owner.id
        }

        db.put_item(self.path, 'certifications', data)


        certification = Certification(id)
        certification._load(data, self.path)

        return certification

    def delete_certification(self, id: str):
        db.delete_item(self.path, 'certifications', id)

    def fetch_certifications(self, **query: typing.Any) -> list[Certification]:
        """
        Récupère une liste de certifications en fonction d'une requête.

        ## Paramètres
        query: `**dict`\n
            La requête pour filtrer les certifications.

        ## Renvoie
        - `list[.Certification]`
        """

        _res = db.fetch('certifications', **query)
        res = []

        for _data in _res:
            cert = Certification()
            cert._load(_data, self.path)

            res.append(cert)

        return res


    # SHORTCUTS

    def get_user(self, id: NSID) -> User | None:
        return self.get_entity(id, User)

    def get_group(self, id: NSID) -> Organization | None:
        return self.get_entity(id, Organization)

    def create_user(self, id: NSID, name: str, position: str = None) -> User:
        return self.create_entity(id, name, User, position)

    def create_group(self, id: NSID, name: str, position: str = None) -> Organization:
        return self.create_entity(id, name, Organization, position)

    def fetch_users(self, **query: typing.Any) -> list[User]:
        """
        Récupère une liste d'utilisateurs en fonction d'une requête.

        ## Paramètres
        query: `**dict`\n
            La requête pour filtrer les utilisateurs.

        ## Renvoie
        - `list[.User]`
        """

        return self.fetch_entities(User, **query)

    def fetch_groups(self, **query: typing.Any) -> list[Organization]:
        """
        Récupère une liste d'organisations en fonction d'une requête.

        ## Paramètres
        query: `**dict`\n
            La requête pour filtrer les organisations.

        ## Renvoie
        - `list[.Organization]`
        """

        return self.fetch_entities(Organization, **query)


import time

from .base import NSID
from .. import database as db


class BankAccount:
    """
    Compte en banque d'une entité, individuelle ou collective.

    ## Attributs
    - id: `NSID`\n
        Identifiant du compte
    - owner: `NSID`\n
        Identifiant du titulaire du compte
    - amount: `int`\n
        Somme d'argent totale sur le compte
    - frozen: `bool`\n
        État gelé ou non du compte
    - bank: `NSID`\n
        Identifiant de la banque qui détient le compte
    - income: `int`\n
        Somme entrante sur le compte depuis la dernière réinitialisation (tous les ~ 28 jours)
    """

    def __init__(self, owner_id: NSID) -> None:
        self._path: str = ""

        self.id: NSID = NSID(owner_id)
        self.owner_id: NSID = NSID(owner_id)
        self.register_date: int = round(time.time())
        self.tag: str = "inconnu"

        self.amount: int = 0
        self.income: int = 0

        self.frozen: bool = False
        self.flagged: bool = False
        self.digicode: str = ''

    def _load(self, _data: dict, path: str) -> None:
        self._path = path

        self.id = NSID(_data['id'])

        self.owner_id = NSID(_data['owner_id'])
        self.register_date = _data['register_date']
        self.tag = _data['tag']

        self.amount = _data['amount']
        self.income = _data['income']

        self.frozen = _data['frozen']
        self.flagged = _data['flagged']
        self.digicode = _data['digicode']

    def _to_dict(self) -> dict:
        return {
            'id': self.id,
            'owner_id': self.owner_id,
            'register_date': self.register_date,
            'tag': self.tag,
            'amount': self.amount,
            'income': self.income,
            'frozen': self.frozen,
            'flagged': self.flagged,
            'digicode': self.digicode
        }

    def save(self):
        db.put_item(self._path, 'accounts', self._to_dict(), True)


    def freeze(self, frozen: bool = True):
        self.frozen = frozen
        self.save()

    def flag(self, flagged: bool = True):
        self.flagged = flagged
        self.save()

    def debit(self, amount: int):
        self.amount -= amount
        self.save()

    def deposit(self, amount: int):
        self.amount += amount
        self.save()


class NSID(str):
    """
    Nation Server ID

    ID unique et universel pour l'ensemble des entités et évènements. Il prend les `int`, les `str` et les autres instances `NSID` pour les convertir en un identifiant hexadécimal.
    """

    unknown = "0"

    admin = "1"
    gov = "2"
    court = "3"
    assembly = "4"

    tresor_public = "A"
    office = "B"
    hexabank = "C"

    def __repr__(self):
        return f"#{self.upper()}"

    def __new__(cls, value):
        if type(value) == int:
            value = hex(value)
        elif type(value) in (str, NSID):
            value = hex(int(value, 16))
        elif value is None:
            value = hex(int(cls.unknown, 16))
        else:
            raise TypeError(f"<{value}> is not NSID serializable")

        if value.startswith("0x"):
            value = value[2:]

        interface = super(NSID, cls).__new__(cls, value.upper())
        return interface

class Interface:
    """
    Instance qui servira de base à toutes les interfaces.
    """

    def __init__(self, path: str):
        self.path = path


import json
import os
import sqlite3
import typing

def get_items(dbpath:str, table: str) -> list[dict[str, typing.Any]]:
	if " " in table:
		raise SyntaxError("Whitespaces are not allowed in <table> parameter")

	with open(os.path.join(dbpath, f"{table}.json")) as _file:
		data = json.load(_file)
		keys = [ k['name'] for k in data['attributes'] ]
		items = data['items']

		return [ 
			dict(zip(keys, item))
			for item in items
		]

def get_item(dbpath:str, table: str, id: str) -> list[str, typing.Any]:
	if " " in table:
		raise SyntaxError("Whitespaces are not allowed in <table> parameter")

	with open(os.path.join(dbpath, f"{table}.json")) as _file:
		data = json.load(_file)

		keys = [ k['name'] for k in data['attributes'] ]
		items = data['items']
		idx = data['idx']

		items = [ 
			dict(zip(keys, item))
			for item in items
		]

	if id in idx:
		return items[idx.index(id)]
	else:
		return


def put_item(dbpath: str, table: str, item: dict, overwrite: bool = False):
	with open(os.path.join(dbpath, f"{table}.json")) as _file:
		data = json.load(_file)

		keys = [ k['name'] for k in data['attributes'] ]
		types = { k['name']: k['type'] for k in data['attributes'] }
		nullables = { k['name']: k['nullable'] for k in data['attributes'] }

		idx = data['idx']
		items = data['items']

		items = [ 
			dict(zip(keys, _item))
			for _item in items
		]


	for k in keys:
		if k not in item.keys():
			raise ValueError(f"Missing key: '{k}'")

	for k, v in item.items():
		if k not in keys:
			raise ValueError(f"Unexpected key: '{k}'")

		if v is None:
			if not nullables[k]:
				raise ValueError(f"Invalid type for key '{k}': Expected {types[k]}, got NoneType")
		elif type(v).__name__ != types[k]:
			raise ValueError(f"Invalid type for key '{k}': Expected {types[k]}, got {type(v).__name__}")


	values = [ item[k] for k in keys ]

	if item['id'] in idx:
		if overwrite:
			items[idx.index(item['id'])] = dict(zip(keys, values))
		else:
			raise ValueError(f"Item with id '{item['id']}' already exists.")
	else:
		items.append(dict(zip(keys, values)))
		idx.append(item['id'])


	with open(os.path.join(dbpath, f"{table}.json"), 'w' , encoding = 'UTF-8') as _file:
		data['items'] = [ list(i.values()) for i in items ]
		data['idx'] = idx

		json.dump(data, _file, indent = 4)


def delete_item(dbpath: str, table: str, id: str) -> tuple[bool, str]:
	if " " in table:
		raise SyntaxError("Whitespaces are not allowed in <table> parameter")

	with open(os.path.join(dbpath, f"{table}.json")) as _file:
		data = json.load(_file)

		keys = [ k['name'] for k in data['attributes'] ]

		idx = data['idx']
		items = data['items']

		items = [ 
			dict(zip(keys, _item))
			for _item in items
		]

	if id in idx:
		items.pop(idx.index(id))
		idx.remove(id)
	else:
		raise KeyError(f"Item with id '{id}' does not exist.")


	with open(os.path.join(dbpath, f"{table}.json"), 'w' , encoding = 'UTF-8') as _file:
		data['items'] = [ i.values() for i in items ]
		data['idx'] = idx

		json.dump(data, _file, indent = 4)

	return True, "Deleted Successfully"

def fetch(dbpath: str, **query: typing.Any) -> list[dict]:
	zone = dbpath.split(':')

	items = sorted(get_items(zone[0], zone[1]), key = lambda item: -int(item.get('register_date', "0")))
	res = []

	for item in items:
		if item is None:
			continue

		for q, value in query.items():	
			try:
				parsed_attr = item[q]

				if parsed_attr != value:
					break
			except (json.JSONDecodeError, KeyError, TypeError):
				if item.get(q) != value:
					break
		else:
			res.append(item)

	return res


import os
import random
import time

from ..models.base import *
from ..models.state import *
from ..models.scale import *

from .. import database as db


class StateInterface(Interface):
    """
    Gère les interactions avec les votes et les officiers.

    ## Informations
    - Liste des partis enregistrés: `.Party`
    - Liste des officiers et candidats: `.Candidate`
    - Résultats des votes: `.Vote`
    """

    def __init__(self, path: str) -> None:
        super().__init__(os.path.join(path, "state"))


    """
    ---- VOTES ----
    """

    def get_vote(self, id: NSID) -> Vote:
        """
        Récupère un vote.

        ## Paramètres
        id: `NSID`\n
            ID du vote.

        ## Renvoie
        - `.Vote`
        """

        id = NSID(id)
        data = db.get_item(self.path, 'votes', id)

        vote = Vote(id)
        vote._load(data, self.path)

        return vote

    def open_vote(
        self,
        title: str = None, # Titre du vote
        author: NSID = 0x0, # ID de l'auteur du vote
        type: str = 'normal', # Type de vote
        anonymous: bool = True, # Vote anonyme ou non
        options: list[dict | str ] = [], # Options du vote
        start: int = round(time.time()), # Début du vote
        end: int = round(time.time() + 3600), # Fin du vote
        min_choices: int = 1, # Nombre minimum de choix
        max_choices: int = 1, # Nombre maximum de choix
        majority: int = 50 # Majorité nécessaire à ce vote
    ) -> Vote:


        opts: dict = {}

        for opt in options:
            if isinstance(opt, str) or opt.get('id') is None:
                _opt_id = NSID(random.randint(100000, 999999))
                _opt_title = opt.get('title', _opt_id) if isinstance(opt, dict) else opt
            else:
                if opt['id'] in opts.keys():
                    raise KeyError(f"Option with key '{opt['id']}' already exists.")
                else:
                    _opt_id = NSID(opt['id'])
                    _opt_title = opt['title']

            opts[_opt_id] = {
                'title': _opt_title,
                'count': 0,
                'voters': []
            }

        _TYPES = (
            'normal',
            'partial',
            'full',
            '2pos',
            '3pos'
        )
        
        if type not in _TYPES:
            raise ValueError(f"Type '{type}' not recognized.")
        
        if min_choices > max_choices:
            raise ValueError("Minimimum choices must be smaller than maximum choices.")
        
        if not 50 <= majority <= 100:
            raise ValueError(f"Majority must be between 50 and 100 both included, not {majority}")
        
        if start - 7200 <= time.time(): # 2h avant le vote pour rajouter des options supplémentaires
            if min_choices > len(opts) or max_choices > len(opts):
                pass


        data = {
            'id': NSID(round(time.time() * 1000)),
            'title': title,
            'author': NSID(author),
            'type': type,
            'anonymous': bool(anonymous),
            'min_choices': int(min_choices),
            'max_choices': int(max_choices),
            'majority': int(majority),
            'start': int(start),
            'end': int(end),
            'options': opts,
            'voters': []
        }

        db.put_item(self.path, 'votes', data)

        vote = Vote(data['id'])
        vote._load(data, self.path)

        return vote

    # Aucune possibilité de supprimer un vote


    """
    PARTIS
    """

    def get_party(self, id: NSID) -> Party:
        """
        Récupère un parti politique.

        ## Paramètres
        id: `NSID`\n
            ID du parti.

        ## Renvoie
        - `.Party`
        """

        id = NSID(id)
        data = db.get_item(self.path, 'parties', str(id))

        if data is None:
            return

        party = Party(id)
        party._load(data, self.path)

        return party

    def register_party(self, id: NSID, color: int, motto: str = None) -> Party:
        """
        Enregistre un nouveau parti pour que ses députés puissent s'y présenter.

        ## Paramètres
        - id: `NSID`\n
            ID de l'entreprise à laquelle correspond le parti
        - color: `int`\n
            Couleur du parti
        - motto: `str`\n
            Devise du parti
        """

        data = {
            'id': NSID(id),
            'color': color,
            'motto': motto
        }

        db.put_item(self.path, 'parties', data)


        # TRAITEMENT

        party = Party(NSID(data['id']))
        party._load(data, self.path)

        return party


    """
    CANDIDATS
    """

    def get_candidate(self, id: NSID) -> Candidate:
        """
        Récupère un candidat.

        ## Paramètres
        id: `NSID`\n
            ID du candidat.

        ## Renvoie
        - `.Candidate`
        """

        id = NSID(id)
        data = db.get_item(self.path, 'candidates', str(id))

        if data is None:
            return

        candidate = Candidate(id)
        candidate._load(data, self.path)

        return candidate


    def add_candidate(self, id: NSID, party: Party = None, scale: dict | Scale = {}) -> Candidate:
        """
        Enregistre un nouveau candidat.

        ## Paramètres
        - id: `NSID`\n
            ID de l'entreprise à laquelle correspond le candidat
        - party: `.Party` (optionnel)\n
            Parti du candidat
        - scale: `.Scale`\n
            Résultats du candidat au test Politiscales
        """

        data = {
            'id': NSID(id),
            'scale': scale._to_dict() if isinstance(scale, Scale) else scale,
            'party': party.id if party else None,
            'current': None,
            'history': {}
        }

        db.put_item(self.path, 'candidates', data)

        candidate = Candidate(NSID(data['id']))
        candidate._load(data, self.path)

        return candidate

    def delete_candidate(self, id: NSID):
        """
        Supprime un candidat.

        ## Paramètres
        - id: `NSID`\n
            ID du candidat.
        """

        id = NSID(id)
        db.delete_item(self.path, 'candidates', str(id))


from __future__ import annotations

import json
import time

from .base import NSID
from .scale import Scale
from .. import database as db


class VoteOption:
    """
    Option disponible lors d'un vote

    ## Attributs
    - title: `str`\n
        Label de l'option
    - count: `int`\n
        Nombre de sympathisants pour cette option
    """

    def __init__(self, title: str):
        self.title: str = title
        self.count: int = 0
        self.voters: list = []

    def __repr__(self) -> dict:
        return json.dumps(self.__dict__)

    def _load(self, _data: dict):
        self.title = str(_data['title'])
        self.count = int(_data['count'])
        self.voters = list(map(NSID, _data['voters']))

class Vote:
    """
    Classe de référence pour les différents votes du serveur

    ## Attributs
    - id: `NSID`\n
        Identifiant du vote
    - title: `str`\n
        Titre du vote
    - author: `NSID`\n
        Identifiant de l'auteur du vote
    - type: `str`\n
        Type du vote (`normal`, `partial`, `full`, `2pos`, `3pos`)
    - majority: `int` de 50 à 100 inclus\n
        Pourcentage nécessaire pour qu'une option soit retenue
    - min_choices: `int`\n
        Nombre minimum de choix possibles
    - max_choices: `int`\n
        Nombre maximum de choix possibles
    - start_date: `int`\n
        Date de début du vote
    - end_date: `int`\n
        Date limite pour voter
    - options: dict[str, .VoteOption]\n
        Liste des choix disponibles
    """

    def __init__(self, id: NSID = None) -> None:
        self._path: str = ''

        self.id: NSID = id if id else NSID(0)
        self.title: str = ""
        self.author: NSID = NSID(0)

        self.anonymous: bool = True
        self.type: str = 'normal'
        
        """
        ## Types de vote

        - Vote normal: `normal`
        - Législatives: `partial`
        - Présidentielles: `full`
        - Pour/Contre: `2pos`
        - Pour/Contre/Blanc: `3pos`
        """

        self.min_choices: int = 1
        self.max_choices: int = 1
        self.majority: int = 50 # Entre 50% et 100% inclus

        self.start_date: int = round(time.time())
        self.end_date: int = 0
        self.voters: list[NSID] = []

        self.options: dict[str, VoteOption] = {}

    def _load(self, _data: dict, path: str) -> None:
        self._path = path

        self.id = NSID(_data['id'])
        self.title = _data['title']
        self.author = NSID(_data['author'])

        self.anonymous = _data['anonymous']
        self.type = _data['type']

        self.max_choices = _data['max_choices']
        self.min_choices = _data['min_choices']
        self.majority = _data['majority']

        self.start_date = _data['start']
        self.end_date = _data['end']
        self.voters = list(map(NSID, _data['voters']))

        self.options = {}

        for _opt_id, opt in _data['options'].items():
            option = VoteOption(opt['title'])
            option._load(opt)

            self.options[_opt_id] = option

    def _to_dict(self) -> dict:
        if self.anonymous:
            for opt in self.options.values():
                opt.voters.clear()

        return {
            'id': self.id,
            'title': self.title,
            'author': self.author,
            'anonymous': self.anonymous,
            'type': self.type,
            'min_choices': self.min_choices,
            'max_choices': self.max_choices,
            'majority': self.majority,
            'start': self.start_date,
            'end': self.end_date,
            'voters': list(map(str, self.voters)),
            'options': { id: opt.__dict__ for id, opt in self.options.items() }
        }

    def save(self):
        db.put_item(self._path, 'votes', self._to_dict(), True)


    def get(self, id: str) -> VoteOption:
        if id in self.options.keys():
            return self.options[id]
        else:
            raise KeyError(f"Option {id} not found in vote {self.id}")

    def add_vote(self, id: str, author: NSID, _save: bool = True):
        """
        Ajoute un vote à l'option spécifiée
        """

        self.get(id).count += 1
        self.get(id).voters.append(author)
        self.voters.append(author)

        if _save:
            self.save()

    def add_votes(self, author: NSID, *ids: str):
        """
        Ajoute un vote aux loptions spécifiées
        """

        for id in ids:
            self.add_vote(id, author, _save = False)

        self.save()

    def close(self):
        """
        Ferme le vote
        """

        self.end_date = round(time.time())
        self.save()


class Party:
    def __init__(self, id: NSID):
        self._path: str = ''

        self.id = id
        self.color: int = 0x000000
        self.motto: str = None

    def _load(self, _data: dict, path: str):
        self._path = path

        self.id = NSID(_data['id'])

        self.color = _data['color']
        self.motto = _data['motto']

    def _to_dict(self) -> dict:
        return {
            'id': self.id,
            'color': self.color,
            'motto': self.motto
        }

    def save(self):
        db.put_item(self._path, 'parties', self._to_dict(), True)


class Candidate:
    def __init__(self, id: NSID):
        self._path: str = ''

        self.id: NSID = id
        self.scale: Scale = Scale()
        self.party: Party = None
        self.current: NSID = None
        self.history: dict = {}

    def _load(self, _data: dict, path: str):
        self._path = path

        self.id = NSID(_data['id'])
        self.scale._load(_data['scale'])

        _party = db.get_item(path, 'parties', _data['party'])

        if _party:
            self.party = Party(NSID(_data['party']))
            self.party._load(_party, path)

        self.current = NSID(_data['current']) if _data['current'] else None
        self.history = _data['history']

    def _to_dict(self) -> dict:
        return {
            'id': self.id,
            'scale': self.scale._to_dict(),
            'party': self.party.id if self.party else None,
            'current': self.current,
            'history': self.history
        }

    def save(self):
        db.put_item(self._path, 'candidates', self._to_dict(), True)


class Scale:
    def __init__(self, _data: dict = None):
        self.democratie: float = 0.0
        self.coertition: float = 0.0
        self.liberte: float = 0.0
        self.integration: float = 0.0
        self.ouverture: float = 0.0
        self.diplomatie: float = 0.0
        self.revolution: float = 0.0

        if _data:
            self._load(_data)

    def _load(self, _data: dict):
        self.democratie = _data.get('DEM', 0.0)
        self.coertition = _data.get('SRV', 0.0)
        self.liberte = _data.get('LIB', 0.0)
        self.integration = _data.get('INT', 0.0)
        self.ouverture = _data.get('MDE', 0.0)
        self.diplomatie = _data.get('PAZ', 0.0)
        self.revolution = _data.get('REV', 0.0)

    def _to_dict(self) -> dict:
        return {
            'DEM': self.democratie,
            'SRV': self.coertition,
            'LIB': self.liberte,
            'INT': self.integration,
            'MDE': self.ouverture,
            'PAZ': self.diplomatie,
            'REV': self.revolution
        }

    def calc_score(self) -> float:
        x = sum({
            'DEM': 12 * self.democratie,
            'SRV': 14 * self.coertition,
            'LIB': 14 * self.liberte,
            'INT': 20 * self.integration,
            'MDE': 12 * self.ouverture, # Ouverture au monde
            'PAZ': 10 * self.diplomatie, # Paix (dans le sens évitement de la guerre)
            'REV': 18 * self.revolution
        }.values())

        return round(x / 40, 2)


REFERENCE = {
    "Anarchie": Scale({
        "DEM": 40,
        "SRV": 40,
        "LIB": 40,
        "INT": 40,
        "MDE": 40,
        "PAZ": 40,
        "REV": 40
    }),
    "Révolution": Scale({
        "DEM": 40,
        "SRV": 25,
        "LIB": 35,
        "INT": 36,
        "MDE": 40,
        "PAZ": 40,
        "REV": 25
    }),
    "Passivité": Scale({
        "DEM": 35,
        "SRV": 10,
        "LIB": 36,
        "INT": 32,
        "MDE": 40,
        "PAZ": 40,
        "REV": 7
    }),
    "Insurrection": Scale({
        "DEM": 30,
        "SRV": 8,
        "LIB": 32,
        "INT": 10,
        "MDE": 20,
        "PAZ": 30,
        "REV": 7
    }),
    "Taoxisme": Scale({
        "DEM": 10,
        "SRV": 5,
        "LIB": 28,
        "INT": 12,
        "MDE": 20,
        "PAZ": 20,
        "REV": 5
    }),
    "Dictature": Scale()
}

def get_positions(score: Scale, ref: dict[str, Scale] = REFERENCE) -> list[str]:
    dist = sorted(
        ref.items(),
        key = lambda r : abs(r[1].calc_score() - score.calc_score())
    )

    items = []

    for item in dist:
        if abs(item[1].calc_score() - score.calc_score()) < 15:
            items.append(item[0])

    return items


import json
import os
import typing

from ..models.base import NSID
from .. import database as db

class Attribute():
	def __init__(self, name: str, _type: typing.Type, nullable: bool = False):
		self.name = name

		if _type.__name__ not in ('NSID', 'str', 'int', 'float', 'bool', 'list', 'dict'):
			raise ValueError(f"Invalid type: {_type}")

		self.type = _type.__name__
		self.nullable = nullable

def gendb(path: str, table: str, attrs: tuple[Attribute]):
	if not os.path.exists(path):
		os.makedirs(path)

	with open(os.path.join(path, f"{table}.json"), 'w') as _db:
		data = {
			'meta': {
				'table_name': table,
			},
			'attributes': [ attr.__dict__ for attr in attrs ],
			'idx': [],
			'items': []
		}

		json.dump(data, _db , indent = 4)

def setup(path: str, include_logs: bool = False, include_drive: bool = False):
	# Drive & Logs

	if include_drive:
		os.makedirs(os.path.join(path, 'drive'), exist_ok = True)

	if include_logs:
		os.makedirs(os.path.join(path, 'entities', 'logs'), exist_ok = True)
		os.makedirs(os.path.join(path, 'economy', 'logs'), exist_ok = True)
		os.makedirs(os.path.join(path, 'state', 'logs'), exist_ok = True)


	# Entités

	gendb(
		path = os.path.join(path, 'entities'),
		table = 'individuals',
		attrs = (
			Attribute('id', NSID),
			Attribute('name', str),
			Attribute('register_date', int),
			Attribute('position', str),
			Attribute('certifications', dict),
			Attribute('additional', dict),
			Attribute('xp', int),
			Attribute('boosts', dict),
		)
	)

	gendb(
		path = os.path.join(path, 'entities'),
		table = 'organizations',
		attrs = (
			Attribute('id', NSID),
			Attribute('name', str),
			Attribute('register_date', int),
			Attribute('position', str),
			Attribute('certifications', dict),
			Attribute('additional', dict),
			Attribute('owner_id', NSID),
			Attribute('members', dict)
		)
	)

	gendb(
		path = os.path.join(path, 'entities'),
		table = 'positions',
		attrs = (
			Attribute('id', str),
			Attribute('name', str),
			Attribute('role', int, nullable = True),
			Attribute('root', str, nullable = True),
			Attribute('level', int, nullable = True),
			Attribute('permissions', dict)
		)
	)

	gendb(
		path = os.path.join(path, 'entities'),
		table = 'certifications',
		attrs = (
			Attribute('id', str),
			Attribute('name', str),
			Attribute('owner', NSID),
			Attribute('parent', NSID),
			Attribute('duration', int)
		)
	)

	db.put_item(os.path.join(path, 'entities'), table = 'positions', item = {
		'id': 'member',
		'name': 'Membre',
		'role': None,
		'root': None,
		'level': None,
		'permissions': {}
	})

	db.put_item(os.path.join(path, 'entities'), table = 'positions', item = {
		'id': 'group',
		'name': 'Groupe',
		'role': None,
		'root': None,
		'level': None,
		'permissions': {}
	})


	# Économie

	gendb(
		path = os.path.join(path, 'economy'),
		table = 'accounts',
		attrs = (
			Attribute('id', NSID),
			Attribute('owner_id', NSID),
			Attribute('register_date', int),
			Attribute('tag', str, nullable = True),
			Attribute('amount', int),
			Attribute('income', int),
			Attribute('frozen', bool),
			Attribute('flagged', bool),
			Attribute('digicode', str)
		)
	)


	# Justice

	gendb(
		path = os.path.join(path, 'state'),
		table = 'reports',
		attrs = (
			Attribute('id', NSID),
			Attribute('target', NSID),
			Attribute('author', NSID),
			Attribute('date', int),
			Attribute('status', int),
			Attribute('reason', str, nullable = True),
			Attribute('details', str, nullable = True)
		)
	)

	gendb(
		path = os.path.join(path, 'state'),
		table = 'lawsuits',
		attrs = (
			Attribute('id', NSID),
			Attribute('target', NSID),
			Attribute('judge', NSID),
			Attribute('title', str),
			Attribute('date', int),
			Attribute('report', NSID, nullable = True),
			Attribute('is_private', bool),
			Attribute('is_open', bool)
		)
	)

	gendb(
		path = os.path.join(path, 'state'),
		table = 'sanctions',
		attrs = (
			Attribute('id', NSID),
			Attribute('target', NSID),
			Attribute('type', str),
			Attribute('date', int),
			Attribute('duration', int),
			Attribute('title', str),
			Attribute('lawsuit', NSID, nullable = True)
		)
	)


	# État & Politique

	gendb(
		path = os.path.join(path, 'state'),
		table = 'votes',
		attrs = (
			Attribute('id', NSID),
			Attribute('title', str),
			Attribute('author', NSID),
			Attribute('type', str),
			Attribute('anonymous', bool),
			Attribute('max_choices', int),
			Attribute('min_choices', int),
			Attribute('majority', int),
			Attribute('start', int),
			Attribute('end', int),
			Attribute('options', dict),
			Attribute('voters', list)
		)
	)

	gendb(
		path = os.path.join(path, 'state'),
		table = 'parties',
		attrs = (
			Attribute('id', NSID),
			Attribute('color', int, nullable = True),
			Attribute('motto', str, nullable = True)
		)
	)

	gendb(
		path = os.path.join(path, 'state'),
		table = 'candidates',
		attrs = (
			Attribute('id', NSID),
			Attribute('scale', dict, nullable = True),
			Attribute('party', NSID, nullable = True),
			Attribute('current', NSID, nullable = True),
			Attribute('history', dict, nullable = True)
		)
	)


import os
import time
import typing

from ..models.base import *
from ..models.economy import *

from .. import database as db, utils


class EconomyInterface(Interface):
	"""Interface qui vous permettra d'interagir avec les comptes en banque et les transactions économiques."""

	def __init__(self, path: str) -> None:
		super().__init__(os.path.join(path, 'economy'))

	"""
	---- COMPTES EN BANQUE ----
	"""

	def get_account(self, id: NSID) -> BankAccount:
		"""
		Récupère les informations d'un compte bancaire.

		## Paramètres
		id: `NSID`\n
			ID du compte.

		## Renvoie
		- `.BankAccount`
		"""

		id = NSID(id)
		
		data = db.get_item(self.path, 'accounts', id)

		if not data:
			return


		# TRAITEMENT

		account = BankAccount(id)
		account._load(data, self.path)

		return account

	def create_account(self, owner: NSID, tag: str = None, flagged: bool = False) -> str:
		"""
		Sauvegarde un compte bancaire dans la base de données.

		## Paramètres
		- account: `.BankAccount`\n
			Compte à sauvegarder
		"""

		res = self.get_account(owner)

		if res:
			id = NSID(round(time.time() * 1000))
		else:
			id = NSID(owner)

		data = {
			'id': id,
			'tag': tag,
			'owner_id': NSID(owner),
			'frozen': False,
			'flagged': flagged,
			'register_date': round(time.time()), 
			'amount': 0,
			'income': 0,
			'digicode': utils.gen_digicode(8)
		}

		db.put_item(self.path, 'accounts', data)


		# TRAITEMENT

		account = BankAccount(owner)
		account._load(data, self.path)

		return account


	def fetch_accounts(self, **query: typing.Any) -> list[BankAccount]:
		"""
		Récupère une liste de comptes en banque en fonction d'une requête.

		## Paramètres
		query: `**dict`\n
			La requête pour filtrer les comptes.

		## Renvoie
		- `list[.BankAccount]`
		"""

		res = db.fetch(f"{self.path}:accounts", **query)


		# TRAITEMENT

		accounts = []

		for _acc in res:
			if not _acc: continue

			account = BankAccount(_acc["owner_id"])

			account.id = NSID(_acc['id'])
			account._load(_acc, self.path)

			accounts.append(account)

		return accounts


import math
import time

EPOCH = 1577833200 # 1er Janvier 2020
PATIENTAGE_DATE = 1725141600 # 1er Septembre 2024
OPEN_DATE = 1756677600 # 1er Septembre 2025

MANDATE_DURATION = 2419200 # 28 jours

def get_cycle(ts: int = round(time.time())):
    if EPOCH <= ts < PATIENTAGE_DATE:
        return 0
    elif PATIENTAGE_DATE <= ts < OPEN_DATE:
        return 1
    elif OPEN_DATE <= ts:
        return math.floor((ts - OPEN_DATE) / MANDATE_DURATION) + 2
    else:
        raise ValueError(f"Timestamp {ts} is out of range (must be greater than or equal to {EPOCH}).")

def get_day(ts: int = round(time.time())) -> int:
    cycle = get_cycle(ts)

    if cycle == 0:
        return math.floor((ts - EPOCH) / 86400)
    elif cycle == 1:
        return math.floor((ts - PATIENTAGE_DATE) / 86400)
    else:
        return math.floor(((ts - OPEN_DATE) % MANDATE_DURATION) / 86400)

def get_phase(ts: int = round(time.time())) -> str:
    cycle = get_cycle(ts)

    if cycle < 2: # Les deux premiers cycles durent (beaucoup) plus longtemps qu'un cycle normal.
        return 'undefined'

    day = get_day(ts)

    if day == 0:
        return 'investiture' # Investiture du PR
    elif 1 <= day < 24:
        return 'paix' # Rien du tout
    elif 24 <= day < 28:
        return 'partials' # Élections législatives
    elif 28 <= day < 52:
        return 'paix' # Toujours rien
    elif 52 <= day < 56:
        return 'elections' # Élections législatives et présidentielles

    else:
        raise ValueError(f"Idk what happened but it seems that {day} is greater than 55...")


def next_election(type: str = 'partial') -> int:
    if get_cycle() == 0:
        return PATIENTAGE_DATE
    elif get_cycle() == 1:
        return OPEN_DATE
    else:
        if type == 'partial':
            ts = OPEN_DATE + get_cycle() * MANDATE_DURATION
        elif type == 'full':
            ts = OPEN_DATE + get_cycle() * MANDATE_DURATION

            if get_cycle() % 2 == 1:
                ts += 28 * 86400

        ts = round(ts / 86400) * 86400

        return ts


import os
import time

from ..models.base import *
from ..models.justice import *

from .. import database as db


class JusticeInterface(Interface):
    """
    Gère les procès, sanctions et signalements.
    """

    def __init__(self, path: str) -> None:
        super().__init__(os.path.join(path, 'state'))

    """
    SIGNALEMENTS
    """

    def get_report(self, id: NSID) -> Report:
        data = db.get_item(self.path, 'reports', id)

        report = Report(id)
        report._load(data, self.path)

        return report

    def submit_report(self, target: NSID, author: NSID, reason: str = None, details: str = None) -> Report:
        data = {
            'id': NSID(round(time.time() * 1000)),
            'target': NSID(target),
            'author': NSID(author),
            'date': round(time.time()),
            'status': 0,
            'reason': reason,
            'details': details
        }

        db.put_item(self.path, 'reports', data)


        # TRAITEMENT

        report = Report(NSID(data['id']))
        report._load(data, self.path)

        return report

    def fetch_reports(self, **query) -> list[Report]:
        res = db.fetch(f"{self.path}:reports", **query)

        reports = []

        for elem in res:
            report = Report(elem['id'])
            report._load(elem, self.path)

            reports.append(report)

        return reports


    """
    PROCÈS
    """

    def get_lawsuit(self, id: NSID) -> Lawsuit:
        data = db.get_item(self.path, 'lawsuits', id)

        lawsuit = Lawsuit(id)
        lawsuit._load(data, self.path)

        return lawsuit

    def open_lawsuit(self, target: NSID, judge: NSID, title: str = None, report: Report = None, private: bool = True) -> Lawsuit:
        data = {
            'id': report.id if report else NSID(round(time.time() * 1000)),
            'target': NSID(target),
            'judge': NSID(judge),
            'title': title,
            'date': round(time.time()),
            'report': report.id if report else None,
            'is_private': private,
            'is_open': False
        }

        db.put_item(self.path, 'lawsuits', data)


        lawsuit = Lawsuit(NSID(data['id']))
        lawsuit._load(data, self.path)

        return lawsuit

    def fetch_lawsuits(self, **query) -> list[Lawsuit]:
        res = db.fetch(f"{self.path}:lawsuits", **query)

        lawsuits = []

        for elem in res:
            lawsuit = Lawsuit(elem['id'])
            lawsuit._load(elem, self.path)

            lawsuits.append(lawsuit)

        return lawsuits


    """
    SANCTIONS
    """

    def get_sanction(self, id: NSID) -> Sanction:
        data = db.get_item(self.path, 'sanctions', self.id)

        sanction = Sanction(id)
        sanction._load(data, self.path)

        return sanction

    def add_sanction(self, target: NSID, _type: str, duration: int = 0, title: str = None, lawsuit: Lawsuit = None) -> Sanction:
        data = {
            'id': lawsuit.id if lawsuit else NSID(round(time.time() * 1000)),
            'target': NSID(target),
            'type': _type,
            'date': round(time.time()),
            'duration': duration,
            'title': title,
            'lawsuit': lawsuit.id if lawsuit else None
        }

        db.put_item(self.path, 'sanctions', data)

        sanction = Sanction(NSID(data['id']))
        sanction._load(data, self.path)

        return sanction

    def fetch_sanctions(self, **query) -> list[Sanction]:
        res = db.fetch(f"{self.path}:sanctions", **query)

        sanctions = []

        for elem in res:
            sanction = Sanction(elem['id'])
            sanction._load(elem, self.path)

            sanctions.append(sanction)

        return sanctions


import os

from .functions.commons import *

def register_file(_dir: str, name: str, data: bytes, overwrite: bool = False):
	path = os.path.join(_dir, name)

	if not os.path.exists(_dir): # La database est bien structurée, aucune création de dossier permise
		raise PermissionError(f"Attempt to create a directory: {path}")

	if os.path.exists(path) and not overwrite:
		raise FileExistsError(f"File {path} already exists.")

	with open(path, "wb") as file:
		file.write(data)

def open_file(path: str) -> bytes:
	path = adjust_path(path)
	if path.startswith('/'): path = path[1:]

	with open(path, "rb") as file:
		return file.read()


from __future__ import annotations

import time
import typing

from ..functions.commons import merge_permissions
from .base import NSID

from .. import database as db

def _load_position(id: str, path: str) -> dict:
    position = db.get_item(path, 'positions', id)

    if position is None:
        return

    if position['root']:
        root = _load_position(position['root'], path)

        p1 = position['permissions']
        p2 = root['permissions']

        position['permissions'] = merge_permissions(p1, p2)

    return position


class PositionPermissions:
    """
    Permissions d'une position à l'échelle du serveur. Certaines sont attribuées selon l'appartenance à divers groupes ayant une position précise
    """

    def __init__(self, _data: list | dict[str, bool] = None) -> None:
        self.create_certifications: bool = False # Créer des certifications
        self.create_entities: bool = False # Créer des entités
        self.create_groups: bool = False # Créer des groupes
        self.create_parties: bool = False # Créer un parti
        self.debit_accounts: bool = False # Débiter les comptes bancaires
        self.edit_constitution: bool = False # Proposer des lois constitutionnelles
        self.edit_laws: bool = False # Proposer des lois
        self.handle_reports: bool = False # Accepter ou refuser les signalements
        self.investigate: bool = False # Accéder aux logs
        self.manage_accounts: bool = False # Gérer les comptes bancaires
        self.manage_bots: bool = False # Gérer la config des bots
        self.manage_certifications: bool = False # Gérer et distribuer les certifications
        self.manage_elections: bool = False # Planifier ou annuler des élections, gérer des candidatures...
        self.manage_entities: bool = False # Gérer les entités
        self.manage_government: bool = False # Créer des minisètres, destituer des ministres, etc.
        self.manage_groups: bool = False # Gérer les groupes
        self.manage_lawsuits: bool = False # Gérer ou ouvrir des poursuites judiciaires
        self.manage_officers: bool = False # Gérer les officiers et agents
        self.manage_parties: bool = False # Gérer les partis
        self.manage_positions: bool = False # Gérer les positions
        self.manage_votes: bool = False # Gérer les élections
        self.moderate_entities: bool = False # Modérer les entités
        self.moderate_groups: bool = False # Modérer les groupes
        self.use_aliases: bool = False # Faire une requête au nom d'une autre entité
        self.vote: bool = False # Voter
        self.vote_laws: bool = False # Voter les lois

        if _data:
            if isinstance(_data, list):
                self.merge(dict(zip(_data, [ True ] * len(_data))))
            elif isinstance(_data, dict):
                self.merge(_data)
            else:
                raise TypeError("Invalid data format for PositionPermissions.")

    def __repr__(self):
        return self.__dict__.__repr__()

    def merge(self, permissions: dict[str, bool] | typing.Self):
        if isinstance(permissions, PositionPermissions):
            permissions: dict[str, bool] = permissions.__dict__

        for key, val in permissions.items():
            perm: bool = self.__getattribute__(key)
            self.__setattr__(key, bool(perm or val))

class Position:
    """
    Position légale d'une entité

    ## Attributs
    - id: `str`\n
        Identifiant de la position
    - name: `str`\n
        Titre de la position
    - root: `str`\n
        Catégorie de la position (officier, citoyen)
    - level: `int`\n
        Grade de la position dans cette catégorie
    - permissions: `.PositionPermissions`\n
        Permissions accordées à l'utilisateur
    """

    def __init__(self, id: str = 'member') -> None:
        self._path: str = ""

        self.id = id
        self.name: str = "Membre"
        self.role: int = None
        self.root: str = None
        self.level: int = None
        self.permissions: PositionPermissions = PositionPermissions()


    def __repr__(self):
        return self.id

    def __eq__(self, value: Position):
        if not isinstance(value, Position):
            return NotImplemented

        return self.id == value.id

    def _lt_(self, value: Position):
        if not isinstance(value, Position):
            return NotImplemented

        a, b = self._compare(value)

        return a < b

    def _le_(self, value: Position):
        if not isinstance(value, Position):
            return NotImplemented

        a, b = self._compare(value)

        return a <= b

    def _gt_(self, value: Position):
        if not isinstance(value, Position):
            return NotImplemented

        a, b = self._compare(value)

        return a > b

    def _ge_(self, value: Position):
        if not isinstance(value, Position):
            return NotImplemented

        a, b = self._compare(value)

        return a >= b

    def _compare(self, value: Position) -> tuple[int, int]:
        if not isinstance(value, Position):
            return NotImplemented

        if self.root == value.root:
            return self.level, value.level
        else:
            _ROOTS = [
                None,
                'member',
                'citizen',
                'officer',
                'admin'
            ]

            _ROOTS_GROUPS = [
                None,
                'group',
                'agency',
                'department'
            ]

            if self.root in _ROOTS:
                if value.root in _ROOTS:
                    return _ROOTS.index(self.root), _ROOTS.index(value.root)
                else:
                    raise ValueError("Cannot compare user and group")
            elif self.root in _ROOTS_GROUPS:
                if value.root in _ROOTS_GROUPS:
                    return _ROOTS_GROUPS.index(self.root), _ROOTS_GROUPS.index(value.root)
                else:
                    raise ValueError("Cannot compare user and group")


    def _load(self, _data: dict, path: str) -> None:
        self._path = path

        self.id = _data['id']
        self.name = _data['name']
        self.role = _data['role']
        self.root = _data['root']
        self.level = _data['level']
        self.permissions.merge(_data['permissions'])

    def _to_dict(self) -> dict:
        return {
            'id': self.id,
            'name': self.name,
            'role': self.role,
            'root': self.root,
            'level': self.level,
            'permissions': self.permissions.__dict__
        }

    def save(self):
        db.put_item(self._path, 'positions', self._to_dict(), True)


class Certification:
    """
    Classe de référence pour les certifications
    """

    def __init__(self, id: NSID) -> None:
        self._path: str = '' # Chemin de la db

        self.id: NSID = NSID(id) # ID hexadécimal de la certification
        self.name: str = "Certification Inconnue"
        self.owner: NSID = NSID() # Entreprise propriétaire de la certification
        self.parent: NSID = NSID() # Certification mère (nécessaire à l'entité qui voudra délivrer celle-ci)
        self.duration: int = 0

    def _load(self, _data: dict, path: str):
        self._path = path

        self.id = NSID(_data['id'])
        self.name = _data['name']
        self.owner = NSID(_data['owner'])
        self.parent = NSID(_data['parent'])
        self.duration = _data['duration']

    def _to_dict(self) -> dict:
        return {
            'id': self.id,
            'name': self.name,
            'owner': self.owner,
            'parent': self.parent,
            'duration': self.duration
        }

    def save(self):
        db.put_item(self._path, 'certifications', self._to_dict(), True)


class Entity:
    """
    Classe de référence pour les entités

    ## Attributs
    - id: `NSID`\n
        Identifiant NSID
    - name: `str`\n
        Nom d'usage
    - register_date: `int`\n
        Date d'enregistrement
    - position: `.Position`\n
        Position civile
    - certifications: `dict`\n
        Titres délivrables par les groupes
    - additional: `dict`\n
        Infos supplémentaires exploitables par différents services
    """

    def __init__(self, id: NSID) -> None:
        self._path: str = '' # Chemin de la db

        self.id: NSID = NSID(id) # ID hexadécimal de l'entité
        self.name: str = "Entité Inconnue"
        self.register_date: int = 0
        self.position: Position = Position()
        self.certifications: dict[NSID, int] = {}
        self.additional: dict = {}

    def __eq__(self, value: Entity):
        if not isinstance(value, Entity):
            return NotImplemented

        return self.id == value.id

    def __lt__(self, value: Entity):
        if not isinstance(value, Entity):
            return NotImplemented

        return self.position < value.position

    def __le__(self, value: Entity):
        if not isinstance(value, Entity):
            return NotImplemented

        return self.position <= value.position

    def __gt__(self, value: Entity):
        if not isinstance(value, Entity):
            return NotImplemented

        return self.position > value.position

    def __ge__(self, value: Entity):
        if not isinstance(value, Entity):
            return NotImplemented

        return self.position >= value.position


    def _load(self, _data: dict, path: str):
        self._path = path

        self.id = NSID(_data['id'])
        self.name = _data['name']
        self.register_date = _data['register_date']
        self.certifications = { NSID(id): exp for id, exp in _data['certifications'].items() }

        position = _load_position(_data['position'], path)
        if position: self.position._load(position, path)

        for  key, value in _data.get('additional', {}).items():
            if isinstance(value, str) and value.startswith('\n'):
                self.additional[key] = int(value[1:])
            else:
                self.additional[key] = value

    def save(self):
        pass

    def set_name(self, name: str) -> None:
        self.name = name
        self.save()

    def set_position(self, position: Position) -> None:
        self.position = position
        self.save()

    def add_certification(self, certification: Certification, __expires: int = 2419200) -> None:
        self.certifications[certification.id] = int(round(time.time()) + __expires)
        self.save()

    def has_certification(self, certification: Certification) -> bool:
        _start = self.certifications.get(certification.id)

        if _start:
            if _start + certification.duration < int(round(time.time())):
                return True
            else:
                self.remove_certification(certification.id)

        return False

    def remove_certification(self, certification: NSID) -> None:
        del self.certifications[certification]
        self.save()

    def add_link(self, key: str, value: str | int) -> None:
        self.additional[key] = value
        self.save()

    def unlink(self, key: str) -> None:
        del self.additional[key]
        self.save()

class User(Entity):
    """
    Entité individuelle

    ## Attributs
    - Tous les attributs de la classe `.Entity`
    - xp: `int`\n
        Points d'expérience de l'entité
    - boosts: `dict[str, int]`\n
        Ensemble des boosts dont bénéficie l'entité
    """

    def __init__(self, id: NSID) -> None:
        super().__init__(NSID(id))

        self.xp: int = 0
        self.boosts: dict[str, int] = {}

    def _load(self, _data: dict, path: str):
        self._path = path

        self.id = NSID(_data['id'])
        self.name = _data['name']
        self.register_date = _data['register_date']
        self.certifications = { NSID(id): exp for id, exp in _data['certifications'].items() }

        position = _load_position(_data['position'], path)
        if position: self.position._load(position, path)

        for  key, value in _data.get('additional', {}).items():
            if isinstance(value, str) and value.startswith('\n'):
                self.additional[key] = int(value[1:])
            else:
                self.additional[key] = value

        self.xp = _data['xp']
        self.boosts = _data['boosts']

    def _to_dict(self) -> dict:
        return {
            'id': self.id,
            'name': self.name,
            'position': self.position.id,
            'register_date': self.register_date,
            'certifications': self.certifications,
            'xp': self.xp,
            'boosts': self.boosts,
            'additional': self.additional
        }

    def save(self):
        db.put_item(self._path, 'individuals', self._to_dict(), True)


    def get_level(self) -> None:
        i = 0
        while self.xp > int(round(25 * (i * 2.5) ** 2, -2)):
            i += 1

        return i

    def add_xp(self, amount: int) -> None:
        boost = 0 if 0 in self.boosts.values() or amount <= 0 else max(list(self.boosts.values()) + [ 1 ])
        self.xp += amount * boost

        self.save()

    def edit_boost(self, name: str, multiplier: int = -1) -> None:
        if multiplier >= 0:
            self.boosts[name] = multiplier
        else:
            del self.boosts[name]

        self.save()

    def get_groups(self) -> list[Entity]:
        res = db.fetch(f"{self._path}:organizations")

        data = []

        for grp in res:
            if grp is None:
                continue

            if grp['owner_id'] == str(self.id):
                data.append(grp)
                continue

            if str(self.id) in grp['members'].keys():
                data.append(grp)
                continue

        groups = []

        for grp in data:
            if grp is None: continue

            group = Organization(NSID(grp['id']))
            group._load(grp, self._path)

            groups.append(group)

        return groups


class GroupMember:
    """
    Membre au sein d'une entité collective

    ## Attributs
    - level: `int`\n
        Niveau d'accréditation d'un membre au sein d'un groupe
    - manager: `bool`\n
        Permission ou non de modifier le groupe
    """

    def __init__(self, id: NSID) -> None:
        self._path: str = ''
        self._group_id: NSID = NSID(0x0)

        self.id = id
        self.level: int = 1 # Plus un level est haut, plus il a de pouvoir sur les autres membres
        self.manager: bool = False

    def __repr__(self):
        return f"level: {self.level}, manager: {self.manager}"

    def __eq__(self, value):
        if not isinstance(value, GroupMember):
            return NotImplemented

        return self.id == value.id

    def __lt__(self, value):
        if not isinstance(value, GroupMember):
            return NotImplemented

        if self.level == value.level:
            return value.manager and not self.manager

        return self.level < value.level

    def __le__(self, value):
        if not isinstance(value, GroupMember):
            return NotImplemented

        if self.level == value.level:
            return value.manager

        return self.level < value.level

    def __gt__(self, value):
        if not isinstance(value, GroupMember):
            return NotImplemented

        if self.level == value.level:
            return self.manager and not value.manager

        return self.level > value.level

    def __ge__(self, value):
        if not isinstance(value, GroupMember):
            return NotImplemented

        if self.level == value.level:
            return self.manager

        return self.level > value.level

    def _load(self, _data: dict, path: str, group: NSID):
        self._path = path
        self._group_id = group

        self.level = _data['level']
        self.manager = _data['manager']

    def _to_dict(self) -> dict:
        return {
            'level': self.level,
            'manager': self.manager
        }

    def save(self):
        data = db.get_item(self._path, 'organizations', self._group_id)

        group = data.copy()
        group['id'] = NSID(group['id'])
        group['owner_id'] = NSID(group['owner_id'])

        group['members'] = {}

        for id, m in data['members'].items():
            if m['level'] > 0:
                group['members'][id] = m

        db.put_item(self._path, 'organizations', group, True)

    def edit(self, level: int = None, manager: bool = None) -> None:
        if level:
            self.level = level
        else:
            return

        if manager is not None:
            self.manager = manager

        self.save()

    def promote(self, level: int = None):
        if level is None:
            level = self.level + 1

        self.edit(level = level)

    def demote(self, level: int = None):
        if level is None:
            level = self.level - 1

        self.edit(level = level)


class Organization(Entity):
    """
    Entité collective

    ## Attributs
    - Tous les attributs de la classe `.Entity`
    - owner: `.Entity`\n
        Utilisateur ou entreprise propriétaire de l'entité collective
    - members: `list[.GroupMember]`\n
        Liste des membres de l'entreprise
    """

    def __init__(self, id: NSID) -> None:
        super().__init__(NSID(id))

        self.owner: Entity = User(NSID(0x0))
        self.avatar_path: str = ''

        self.members: dict[NSID, GroupMember] = {}

    def _load(self, _data: dict, path: str):
        self._path = path

        self.id = NSID(_data['id'])
        self.name = _data['name']
        self.register_date = _data['register_date']
        self.certifications = { NSID(id): exp for id, exp in _data['certifications'].items() }

        position = _load_position(_data['position'], path)
        if position: self.position._load(position, path)

        for  key, value in _data.get('additional', {}).items():
            if isinstance(value, str) and value.startswith('\n'):
                self.additional[key] = int(value[1:])
            else:
                self.additional[key] = value


        _owner = db.get_item(path, 'individuals', _data['owner_id'])
        _class = 'user'

        if _owner is None:
            _owner = db.get_item(path, 'organizations', _data['owner_id'])
            _class = 'group'

        if _owner:
            if _class == 'user':
                self.owner = User(_owner['id'])
            elif _class == 'group':
                self.owner = Organization(_owner['id'])
            else:
                self.owner = Entity(_owner['id'])

            self.owner._load(_owner, path)
        else:
            self.owner = None

        for _id, _member in _data['members'].items():
            member = GroupMember(NSID(_id))
            member._load(_member, path, self.id)

            self.members[NSID(member.id)] = member

    def _to_dict(self) -> dict:
        return {
            'id': self.id,
            'name': self.name,
            'position': self.position.id,
            'register_date': self.register_date,
            'owner_id': self.owner.id,
            'members': { id: member._to_dict() for id, member in self.members.items() },
            'certifications': self.certifications,
            'additional': self.additional
        }

    def save(self):
        db.put_item(self._path, 'organizations', self._to_dict(), True)

    def add_member(self, member: NSID) -> GroupMember:
        if not isinstance(member, NSID):
            raise TypeError("L'entrée membre doit être de type NSID")

        member = GroupMember(member)
        member._group_id = self.id
        member._path = self._path

        self.members[member.id] = member

        self.save()
        return member

    def remove_member(self, member: GroupMember) -> None:
        member.demote(level = 0)

    def set_owner(self, member: User) -> None:
        self.owner = member
        self.save()

    def get_member(self, id: NSID) -> GroupMember:
        return self.members.get(id)

    def get_members_by_attr(self, attribute: str = "id") -> list[str]:
        return [ member.__getattribute__(attribute) for member in self.members.values() ]

    def save_avatar(self, data: bytes = None):
        pass